#!/usr/bin/env python
# Created by "Thieu" at 16:10, 31/07/2023 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

import csv
from pathlib import Path


def write_dict_to_csv(data:dict, save_path=None, file_name=None):
    """
    Write a list of dictionaries to a CSV file.

    Args:
        data (list): A list of dictionaries.
        save_path (str): Path to save the file
        file_name (str): The name of the output CSV file.

    Returns:
        None
    """
    save_file = f"{save_path}/{file_name}.csv"
    Path(save_path).mkdir(parents=True, exist_ok=True)
    is_file_existed = Path(save_file).exists()

    with open(save_file, mode='a', newline='') as file:
        writer = csv.DictWriter(file, fieldnames=data.keys())
        if not is_file_existed:
            writer.writeheader()
        writer.writerow(data)
    return None



#!/usr/bin/env python
# Created by "Thieu" at 14:15, 25/05/2023 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

import numpy as np
from metacluster import get_dataset

np.random.seed(41)


def test_Data_class():
    data = get_dataset("circles")
    data.X, scaler = data.scale(data.X, method="MinMaxScaler", feature_range=(0, 1))

    assert 0 <= data.X[np.random.randint(0, 3)][0] <= 1



#!/usr/bin/env python
# Created by "Thieu" at 17:39, 30/07/2023 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
from sklearn import metrics
import numpy as np


### ELBOW
def get_clusters_by_elbow(X, list_clusters=None, **kwargs):
    """
    1. First, apply K-means clustering to the dataset for a range of different values of K, where K is the number of clusters. For example, you might try K=1,2,3,...,10.
    2. For each value of K, compute the Sum of Squared Errors (SSE), which is the sum of the squared distances between each data point and its assigned centroid. The SSE can be obtained from the KMeans object's 'inertia_' attribute.
    3. Plot the SSE for each value of K. You should see that the SSE decreases as K increases, because as K increases, the centroids are closer to the data points. However, at some point, increasing K further will not improve the SSE as much. The idea of the elbow method is to identify the value of K at which the SSE starts to level off or decrease less rapidly, forming an "elbow" in the plot. This value of K is considered the optimal number of clusters.
    """
    if type(list_clusters) in (list, tuple, np.ndarray):
        list_clusters = [item for item in list_clusters]
    else:
        list_clusters = list(range(2, 15))
    wcss = []
    for n_c in list_clusters:
        kmeans = KMeans(n_clusters=n_c)
        kmeans.fit(X=X)
        wcss.append(kmeans.inertia_)
    x1, y1 = 2, wcss[0]
    x2, y2 = list_clusters[-1], wcss[-1]
    distances = []
    for i in range(len(wcss)):
        x0 = i + 2
        y0 = wcss[i]
        numerator = abs((y2 - y1) * x0 - (x2 - x1) * y0 + x2 * y1 - y2 * x1)
        denominator = np.sqrt((y2 - y1) ** 2 + (x2 - x1) ** 2)
        distances.append(numerator / denominator)
    return distances.index(max(distances)) + 2
### END ELBOW


### gap statistics
def get_clusters_by_gap_statistic(X, list_clusters=None, B=10, N_init=10, **kwargs):
    gaps, s_k, K = compute_gap_statistic(X, refs=None, B=B, K=list_clusters, N_init=N_init)

    gaps_thres = gaps - s_k
    below_or_above = (gaps[0:-1] >= gaps_thres[1:])
    if below_or_above.any():
        optimal_k = K[below_or_above.argmax()]
    else:
        optimal_k = K[-1]
    return optimal_k


def compute_Wk(data, classification_result):
    '''
    This function computes the Wk after each clustering

    data:np.array, containing all the data
    classification_result: np.array, containing all the clustering results for all the data
    '''
    Wk = 0
    label_set = set(classification_result.tolist())
    for label in label_set:
        each_cluster = data[classification_result == label, :]
        mu = each_cluster.mean(axis=0)
        D = sum(sum((each_cluster - mu) ** 2)) * 2.0 * each_cluster.shape[0]
        Wk = Wk + D / (2.0 * each_cluster.shape[0])
    return Wk


def compute_gap_statistic(X, refs=None, B=10, K=range(2, 11), N_init=10):
    """
    This function first generates B reference samples; for each sample, the sample size is the same as the original datasets;
    the value for each reference sample follows a uniform distribution for the range of each feature of the original datasets;
    using a simplify formula to compute the D of each cluster, and then the Wk; K should be a increment list, 1-10 is fair enough;
    the B value is about the number of replicated samples to run gap-statistics, it is recommended as 10, and it should not be changed/decreased that to a smaller value;

    Parameters
    ----------
    X :  np.array, the original data;
    refs : np.array or None, it is the replicated data that you want to compare with if there exists one; if no existing replicated/proper data, just use None, and the function  will automatically generates them;
    B : int, the number of replicated samples to run gap-statistics; it is recommended as 10, and it should not be changed/decreased that to a smaller value;
    K : list type, the range of K values to test on;
    N_init : int, states the number of initial starting points for each K-mean running under sklearn, in order to get stable clustering result each time;

    Returns
    -------
        gaps: np.array, containing all the gap-statistics results;
        s_k: float, the baseline value to minus with; say reference paper for detailed meaning;
        K: list, containing all the tested K values;
    """
    shape = X.shape
    if refs == None:
        tops = X.max(axis=0)
        bots = X.min(axis=0)
        dists = np.matrix(np.diag(tops - bots))
        rands = np.random.random_sample(size=(shape[0], shape[1], B))
        for i in range(B):
            rands[:, :, i] = rands[:, :, i] * dists + bots
    else:
        rands = refs

    gaps = np.zeros(len(K))
    Wks = np.zeros(len(K))
    Wkbs = np.zeros((len(K), B))

    for indk, k in enumerate(K):
        # n_init is the number of times each Kmeans running to get stable clustering results under each K value
        k_means = KMeans(n_clusters=k, init='k-means++', n_init=N_init, max_iter=300, tol=0.0001, verbose=0, random_state=None, copy_x=True)
        k_means.fit(X)
        classification_result = k_means.labels_
        # compute the Wk for the classification result
        Wks[indk] = compute_Wk(X, classification_result)

        # clustering on B reference datasets for each 'k'
        for i in range(B):
            Xb = rands[:, :, i]
            k_means.fit(Xb)
            classification_result_b = k_means.labels_
            Wkbs[indk, i] = compute_Wk(Xb, classification_result_b)

    # compute gaps and sk
    gaps = (np.log(Wkbs)).mean(axis=1) - np.log(Wks)
    sd_ks = np.std(np.log(Wkbs), axis=1)
    sk = sd_ks * np.sqrt(1 + 1.0 / B)
    return gaps, sk, K


### silhouette score
def get_clusters_by_silhouette_score(X, list_clusters=None, **kwargs):
    if type(list_clusters) in (list, tuple, np.ndarray):
        list_clusters = [item for item in list_clusters]
    else:
        list_clusters = list(range(2, 15))
    sil_max = 0
    sil_max_clusters = 2
    for n_clusters in list_clusters:
        model = KMeans(n_clusters=n_clusters)
        labels = model.fit_predict(X)
        sil_score = metrics.silhouette_score(X, labels)
        if sil_score > sil_max:
            sil_max = sil_score
            sil_max_clusters = n_clusters
    return sil_max_clusters
### END silhouette score


### DB score
def get_clusters_by_davies_bouldin(X, list_clusters=None, **kwargs):
    if type(list_clusters) in (list, tuple, np.ndarray):
        list_clusters = [item for item in list_clusters]
    else:
        list_clusters = list(range(2, 15))
    db_min = np.finfo(np.float64)
    db_min_clusters = 2
    for n_clusters in list_clusters:
        model = KMeans(n_clusters=n_clusters)
        labels = model.fit_predict(X)
        db_score = metrics.davies_bouldin_score(X, labels)
        if db_score < db_min:
            db_min = db_score
            db_min_clusters = n_clusters
    return db_min_clusters
### END DB score


### Calinski-Harabasz Index
def get_clusters_by_calinski_harabasz(X, list_clusters=None, **kwargs):
    if type(list_clusters) in (list, tuple, np.ndarray):
        list_clusters = [item for item in list_clusters]
    else:
        list_clusters = list(range(2, 15))
    ch_max = 0
    ch_max_clusters = 2
    for n_clusters in list_clusters:
        model = KMeans(n_clusters=n_clusters)
        labels = model.fit_predict(X)
        ch_score = metrics.calinski_harabaz_score(X, labels)
        if ch_score > ch_max:
            ch_max = ch_score
            ch_max_clusters = n_clusters
    return ch_max_clusters
### END Calinski-Harabasz Index


### Bayesian Information Criterion
def get_clusters_by_bic(X, list_clusters=None, **kwargs):
    if type(list_clusters) in (list, tuple, np.ndarray):
        list_clusters = [item for item in list_clusters]
    else:
        list_clusters = list(range(2, 15))
    bic_max = 0
    bic_max_clusters = 2
    for n_clusters in list_clusters:
        gm = GaussianMixture(n_components=n_clusters, n_init=10, tol=1e-3, max_iter=1000).fit(X)
        bic_score = -gm.bic(X)
        if bic_score > bic_max:
            bic_max = bic_score
            bic_max_clusters = n_clusters
    return bic_max_clusters
### END Bayesian Information Criterion


def get_clusters_all_min(X, list_clusters=None, **kwargs):
    k_list = compute_all_methods(X, list_clusters, **kwargs)
    return min(k_list)


def get_clusters_all_max(X, list_clusters=None, **kwargs):
    k_list = compute_all_methods(X, list_clusters, **kwargs)
    return max(k_list)


def get_clusters_all_mean(X, list_clusters=None, **kwargs):
    k_list = compute_all_methods(X, list_clusters, **kwargs)
    return int(np.mean(k_list))


def get_clusters_all_majority(X, list_clusters=None, **kwargs):
    k_list = compute_all_methods(X, list_clusters, **kwargs)
    return max(set(k_list), key=k_list.count)


def compute_all_methods(X, list_clusters=None, **kwargs):
    k1 = get_clusters_by_elbow(X, list_clusters, **kwargs)
    k2 = get_clusters_by_gap_statistic(X, list_clusters, **kwargs)
    k3 = get_clusters_by_silhouette_score(X, list_clusters, **kwargs)
    k4 = get_clusters_by_davies_bouldin(X, list_clusters, **kwargs)
    k5 = get_clusters_by_calinski_harabasz(X, list_clusters, **kwargs)
    k6 = get_clusters_by_bic(X, list_clusters, **kwargs)
    return [k1, k2, k3, k4, k5, k6]



#!/usr/bin/env python
# Created by "Thieu" at 17:17, 31/07/2023 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

import pandas as pd
import plotly.express as px
import plotly.io as pio

pio.kaleido.scope.mathjax = None


def export_boxplot_figures(df, xlabel="Optimizer", ylabel=None, title="Boxplot of comparison models",
                           show_legend=True, show_mean_only=False,
                           exts=(".png", ".pdf"), file_name="boxplot", save_path="history"):
    if xlabel is None:
        xlabel = ""
    if ylabel is None:
        ylabel = ""
    if title is None:
        title = ""
    boxmean = True if show_mean_only else "sd"
    col_name = list(df.columns)[-1]
    if ylabel is None:
        ylabel = col_name
    fig = px.box(df, x="optimizer", y=col_name, color="optimizer", labels={"model": xlabel, col_name: ylabel})
    fig.update_traces(boxmean=boxmean)  # boxmean=True if want to show mean only
    fig.update_layout(margin=dict(l=20, r=20, t=30, b=20), showlegend=show_legend,
                      title={'text': title, 'x': 0.5, 'xanchor': 'center', 'yanchor': 'top'})
    for idx, ext in enumerate(exts):
        fig.write_image(f"{save_path}/{file_name}{ext}")


def export_convergence_figures(df, xlabel="Epoch", ylabel="Fitness value",
                               title="Convergence chart of comparison models",
                               exts=(".png", ".pdf"), file_name="boxplot", save_path="history"):
    if xlabel is None:
        xlabel = ""
    if ylabel is None:
        ylabel = ""
    if title is None:
        title = ""
    # Melt the DataFrame to convert it from wide to long format
    df = df.reset_index()
    df_long = pd.melt(df, id_vars='index', var_name='Column', value_name='Value')
    # Define the line chart using Plotly Express
    fig = px.line(df_long, x='index', y='Value', color='Column', labels={'index': xlabel, 'Value': ylabel, 'Column': 'Optimizer'})
    fig.update_layout(margin=dict(l=20, r=20, t=30, b=20), showlegend=True,
                      title={'text': title, 'x': 0.5, 'xanchor': 'center', 'yanchor': 'top'})
    for idx, ext in enumerate(exts):
        fig.write_image(f"{save_path}/{file_name}{ext}")



#!/usr/bin/env python
# Created by "Thieu" at 05:36, 28/07/2023 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

import time
from pathlib import Path
import pandas as pd
import numpy as np
from metacluster.utils import mealpy_util as mu, cluster
from metacluster.utils import validator
from metacluster.utils.io_util import write_dict_to_csv
from metacluster.utils.mealpy_util import get_all_optimizers, KCenterClusteringProblem
from metacluster.utils.visualize_util import export_boxplot_figures, export_convergence_figures


class MetaCluster:
    """
    Defines a MetaCluster class that hold all Metaheuristic-based K-Center Clustering methods

    Parameters
    ----------

    list_optimizer=None, list_paras=None, list_obj=None, n_trials=5

    list_optimizer: list, default = None
        List of strings that represent class optimizer or list of instance of Optimizer class from Mealpy library.
        Current supported optimizers, please check it here: https://github.com/thieu1995/mealpy
        If a custom optimizer is passed, make sure it is an instance of `Optimizer` class.
        Please use this to get supported optimizers: MetaCluster.get_support(name="optimizer")

    list_paras: list, default=None
        List of dictionaries that present the parameters of each Optimizer class.
        You can set it to None to use all of default parameters in Mealpy library.

    list_obj : list, default=None
        List of strings that represent objective name.
        Current supported objectives, please check it here: https://github.com/thieu1995/permetrics
        Please use this to get supported objectives: MetaCluster.get_support(name="obj")

    n_trials : int, default=5
        The number of runs for each optimizer for each objective

    Examples
    --------
    The following example shows how to use  the most informative features in the MhaSelector FS method

    >>> from metacluster import get_dataset, MetaCluster
    >>> from sklearn.preprocessing import MinMaxScaler
    >>>
    >>> scaler = MinMaxScaler(feature_range=(0, 1))
    >>> data = get_dataset("aniso")
    >>> data.X = scaler.fit_transform(data.X)
    >>>
    >>> # Get all supported methods and print them out
    >>> MetaCluster.get_support(name="all")
    >>>
    >>> list_optimizer = ["BaseFBIO", "OriginalGWO", "OriginalSMA"]
    >>> list_paras = [
    >>>    {"name": "FBIO", "epoch": 10, "pop_size": 30},
    >>>    {"name": "GWO", "epoch": 10, "pop_size": 30},
    >>>    {"name": "SMA", "epoch": 10, "pop_size": 30}
    >>> ]
    >>> list_obj = ["BHI", "MIS", "XBI"]
    >>> list_metric = ["BRI", "DBI", "DRI", "DI", "KDI"]
    >>> model = MetaCluster(list_optimizer=list_optimizer, list_paras=list_paras, list_obj=list_obj, n_trials=3)
    >>> model.execute(data=data, cluster_finder="elbow", list_metric=list_metric, save_path="history", verbose=False)
    """

    SUPPORT = {
        "cluster_finder": {"elbow": "get_clusters_by_elbow", "gap": "get_clusters_by_gap_statistic",
                           "silhouette": "get_clusters_by_silhouette_score", "davies_bouldin": "get_clusters_by_davies_bouldin",
                           "calinski_harabasz": "get_clusters_by_calinski_harabasz", "bic": "get_clusters_by_bic",
                           "all_min": "get_clusters_by_all_min", "all_max": "get_clusters_by_all_max",
                           "all_mean": "get_clusters_by_all_mean", "all_majority": "get_clusters_by_all_majority"},
        "obj": {"BHI": "min", "CHI": "max", "XBI": "min", "BRI": "max", "DBI": "min", "DRI": "min",
                "DI": "max", "KDI": "min", "LDRI": "min", "LSRI": "min", "SI": "max", "MIS": "max",
                "NMIS": "max", "RaS": "max", "FMS": "max", "HS": "max", "CS": "max", "VMS": "max",
                "PrS": "max", "ReS": "max", "FmS": "max", "CDS": "max", "HGS": "max", "JS": "max",
                "KS": "max", "MNS": "min", "PhS": "max", "RTS": "max", "RRS": "max", "SS1S": "max", "SS2S": "max"},
        "optimizer": list(get_all_optimizers().keys())
    }

    FILENAME_LABELS = "result_labels"
    FILENAME_METRICS = "result_metrics"
    FILENAME_METRICS_MEAN = "result_metrics_min"
    FILENAME_METRICS_STD = "result_metrics_mean"
    FILENAME_CONVERGENCES = "result_convergences"
    HYPHEN_SYMBOL = "="

    def __init__(self, list_optimizer=None, list_paras=None, list_obj=None, n_trials=5):
        self.list_optimizer, self.list_paras = self._set_list_optimizer(list_optimizer, list_paras)
        self.list_obj = self._set_list_function(list_obj, name="objectives")
        self.n_trials = n_trials

    @staticmethod
    def get_support(name="all", verbose=True):
        if name == "all":
            if verbose:
                for key, value in MetaCluster.SUPPORT.items():
                    print(f"Supported methods for '{key}' are: ")
                    print(value)
            return MetaCluster.SUPPORT
        if name in list(MetaCluster.SUPPORT.keys()):
            if verbose:
                print(f"Supported methods for '{name}' are: ")
                print(MetaCluster.SUPPORT[name])
            return MetaCluster.SUPPORT[name]
        raise ValueError(f"MetaCluster doesn't support {name}.")

    def _set_list_function(self, list_obj=None, name="objectives"):
        if type(list_obj) in (list, tuple, np.ndarray):
            list_obj1 = []
            list_obj0 = []
            for obj in list_obj:
                if obj in list(self.SUPPORT["obj"].keys()):
                    list_obj1.append(obj)
                else:
                    list_obj0.append(obj)
            if len(list_obj0) > 0:
                print(f"MetaCluster doesn't support {name}: {list_obj0}")
            return list_obj1

    def _set_list_optimizer(self, list_optimizer=None, list_paras=None):
        if type(list_optimizer) not in (list, tuple):
            raise ValueError("list_optimizers should be a list or tuple.")
        else:
            if list_paras is None or type(list_paras) not in (list, tuple):
                list_paras = [{}, ] * len(list_optimizer)
            elif len(list_paras) != len(list_optimizer):
                raise ValueError("list_paras should be a list with the same length as list_optimizer")
            list_opts = []
            for idx, opt in enumerate(list_optimizer):
                if type(opt) is str:
                    opt_class = mu.get_optimizer_by_name(opt)
                    if type(list_paras[idx]) is dict:
                        list_opts.append(opt_class(**list_paras[idx]))
                    else:
                        list_opts.append(opt_class(epoch=100, pop_size=30))
                elif isinstance(opt, mu.Optimizer):
                    if type(list_paras[idx]) is dict:
                        opt.set_parameters(list_paras[idx])
                    list_opts.append(opt)
                else:
                    raise TypeError(f"optimizer needs to set as a string and supported by Mealpy library.")
        return list_opts, list_paras

    def __run__(self, model, problem, mode="single", n_workers=2, termination=None):
        best_position, best_fitness = model.solve(problem, mode=mode, n_workers=n_workers, termination=termination)
        return {
            "best_fitness": best_fitness,
            "best_solution": best_position,
            "convergence": model.history.list_global_best_fit
        }

    def execute(self, data=None, cluster_finder="elbow", list_metric=None, save_path="history", save_figures=True,
            verbose=True, mode='single', n_workers=None, termination=None):
        """
        Parameters
        ----------
        data : instance of Data class, default=None
            The instance of Data class, make sure you have at least matrix feature X. Or target labels y (Optional).
            Also make sure your matrix X is normalized or standardized

        cluster_finder : str, default="elbow".
            The method to find the optimal number of clusters in data

        list_metric : list, default=None
            List of performance metrics that supported by the library: https://github.com/thieu1995/permetrics
            To get the supported metrics, please use: MetaCluster.get_support(), supported obj are supported metrics

        save_path : str, default="history"
            The path to the folder that hold results

        save_figures : bool, default=True
            Save all of available figure or not

        verbose : int, default = True
            Controls verbosity of output for each training process of each optimizer.

        mode : str, default = 'single'
            The mode used in Optimizer belongs to Mealpy library. Parallel: 'process', 'thread'; Sequential: 'swarm', 'single'.

                - 'process': The parallel mode with multiple cores run the tasks
                - 'thread': The parallel mode with multiple threads run the tasks
                - 'swarm': The sequential mode that no effect on updating phase of other agents
                - 'single': The sequential mode that effect on updating phase of other agents, default

        n_workers : int or None, default = None
            The number of workers (cores or threads) used in Optimizer (effect only on parallel mode)

        termination : dict or None, default = None
            The termination dictionary or an instance of Termination class. It is for Optimizer belongs to Mealpy library.
        """
        if data.y is not None:
            n_clusters = len(np.unique(data.y))
        else:
            cluster_finder = validator.check_str("cluster_finder", cluster_finder, list(self.SUPPORT["cluster_finder"].keys()))
            n_clusters = getattr(cluster, cluster_finder)(data.X)
        lb = np.min(data.X, axis=0).tolist() * n_clusters
        ub = np.max(data.X, axis=0).tolist() * n_clusters
        obj_paras = {"decimal": 4}
        log_to = "console" if verbose else "None"
        list_metric = self._set_list_function(list_metric, name="metrics")

        ## Check parent directories
        save_path = f"{save_path}/{data.get_name()}"
        Path(save_path).mkdir(parents=True, exist_ok=True)

        list_problems = []
        for idx_opt, opt in enumerate(self.list_optimizer):
            for idx_obj, obj in enumerate(self.list_obj):

                list_dict = []
                for idx_trial, trial in enumerate(range(self.n_trials)):
                    print(f"MetaCluster are working on: optimizer={opt.get_name()}, obj={obj}, trial={trial+1}")
                    minmax = self.SUPPORT["obj"][obj]
                    prob = KCenterClusteringProblem(lb, ub, minmax, data=data, obj_name=obj, obj_paras=obj_paras, log_to=log_to)
                    list_problems.append(prob)

                    time_run = time.perf_counter()
                    res = self.__run__(opt, prob, mode=mode, n_workers=n_workers, termination=termination)
                    time_run = round(time_run, 5)
                    y_pred = prob.get_y_pred(data.X, res["best_solution"])
                    y_pred = self.HYPHEN_SYMBOL.join(map(str, y_pred))     # Convert all labels to single string to save to csv file.
                    conv = self.HYPHEN_SYMBOL.join(map(str, res["convergence"]))
                    dict_metrics = prob.get_metrics(res["best_solution"], list_metric)

                    ## Save result_labels.csv file
                    dict1 = {"optimizer": opt.get_name(), "obj": obj, "n_clusters": n_clusters, "y_pred": y_pred}
                    write_dict_to_csv(dict1, save_path=save_path, file_name=self.FILENAME_LABELS)

                    ## Save result_metrics.csv file
                    dict2 = {"optimizer": opt.get_name(), "obj": obj, "trial": trial+1, "n_clusters": n_clusters, "time_run": time_run}
                    dict3 = {**dict2, **dict_metrics}
                    write_dict_to_csv(dict3, save_path=save_path, file_name=self.FILENAME_METRICS)

                    ## Save results for metrics-min and metrics-std
                    dict4 = {"time_run": time_run, **dict_metrics}
                    list_dict.append(dict4)

                    ## Save result_convergence.csv
                    dict5 = {"optimizer": opt.get_name(), "obj": obj, "trial": trial+1, "n_clusters": n_clusters, "fitness": conv}
                    write_dict_to_csv(dict5, save_path=save_path, file_name=self.FILENAME_CONVERGENCES)

                ## Save result_metrics_min.csv and result_metrics_std.csv file
                df0 = pd.DataFrame(list_dict)
                dict_mean = df0.mean().to_dict()
                dict_std = df0.std().to_dict()
                dict_mean = {"optimizer": opt.get_name(), "obj": obj, "n_clusters": n_clusters, **dict_mean}
                dict_std = {"optimizer": opt.get_name(), "obj": obj, "n_clusters": n_clusters, **dict_std}
                write_dict_to_csv(dict_mean, save_path=save_path, file_name=self.FILENAME_METRICS_MEAN)
                write_dict_to_csv(dict_std, save_path=save_path, file_name=self.FILENAME_METRICS_STD)

        if save_figures:
            for idx_metric, metric in enumerate(list_metric):
                df = pd.read_csv(f"{save_path}/{self.FILENAME_METRICS}.csv", usecols=["optimizer", "obj", metric])
                for idx_obj, obj in enumerate(self.list_obj):
                    df_draw = df[df["obj"] == obj][["optimizer", metric]]
                    export_boxplot_figures(df_draw, file_name=f"boxplot-{obj}-{metric}", save_path=save_path)

            df = pd.read_csv(f"{save_path}/{self.FILENAME_CONVERGENCES}.csv", usecols=["optimizer", "obj", "trial", "fitness"])
            for idx_obj, obj in enumerate(self.list_obj):
                ## Draw convergence for single trial
                for idx_trial, trial in enumerate(range(self.n_trials)):
                    df_draw = df[(df["obj"] == obj) & (df["trial"] == trial+1)][["optimizer", "fitness"]]
                    df_draw.set_index("optimizer", inplace=True)
                    dict_draw = df_draw.to_dict()["fitness"]
                    for key, value in dict_draw.items():
                        dict_draw[key] = np.array(value.split(self.HYPHEN_SYMBOL), dtype=float)
                    df_draw = pd.DataFrame(dict_draw)
                    export_convergence_figures(df_draw, file_name=f"convergence-{obj}-{trial+1}", save_path=save_path)

                ## Draw mean convergence of all trials
                df_draw = df[df["obj"] == obj][["optimizer", 'fitness']]
                mylist = df_draw.values.tolist()
                dict_mean = {}
                for idx, item in enumerate(mylist):
                    if item[0] in dict_mean:
                        dict_mean[item[0]].append(np.array(item[1].split(self.HYPHEN_SYMBOL), dtype=float))
                    else:
                        dict_mean[item[0]] = [np.array(item[1].split(self.HYPHEN_SYMBOL), dtype=float)]
                for key, value in dict_mean.items():
                    dict_mean[key] = np.mean(value, axis=0)
                df_draw = pd.DataFrame(dict_mean)
                export_convergence_figures(df_draw, file_name=f"convergence-{obj}-mean", save_path=save_path)



#!/usr/bin/env python
# Created by "Thieu" at 15:23, 06/03/2022 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%
#
# from metacluster import get_dataset, MetaCluster
# from sklearn.preprocessing import MinMaxScaler
#
# # Get all supported methods and print them out
# MetaCluster.get_support(name="all")
#
# # Scale dataset to range (0, 1)
# scaler = MinMaxScaler(feature_range=(0, 1))
# data = get_dataset("aniso")
# data.X = scaler.fit_transform(data.X)
#
# # Set up Metaheuristic Algorithms
# list_optimizer = ["BaseFBIO", "OriginalGWO", "OriginalSMA"]
# list_paras = [
#     {"name": "FBIO", "epoch": 10, "pop_size": 30},
#     {"name": "GWO", "epoch": 10, "pop_size": 30},
#     {"name": "SMA", "epoch": 10, "pop_size": 30}
# ]
#
# # Set up list objectives and list performance metrics
# list_obj = ["BHI", "MIS", "XBI"]
# list_metric = ["BRI", "DBI", "DRI", "DI", "KDI"]
#
# # Define MetaCluster model and execute it
# model = MetaCluster(list_optimizer=list_optimizer, list_paras=list_paras, list_obj=list_obj, n_trials=3)
# model.execute(data=data, cluster_finder="elbow", list_metric=list_metric, save_path="history", verbose=False)

__version__ = "0.2.0"

from metacluster.utils.encoder import LabelEncoder
from metacluster.utils.data_loader import Data, get_dataset
from metacluster.metacluster import MetaCluster



#!/usr/bin/env python
# Created by "Thieu" at 17:18, 31/05/2023 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

import numpy as np
from metacluster import get_dataset, MetaCluster

np.random.seed(42)


def test_MetaCluster_class():
    data = get_dataset("circles")
    data.X, scaler = data.scale(data.X, method="MinMaxScaler", feature_range=(0, 1))

    list_optimizer = ["BaseFBIO", "OriginalGWO", "OriginalSMA"]
    list_paras = [
        {"name": "FBIO", "epoch": 10, "pop_size": 30},
        {"name": "GWO", "epoch": 10, "pop_size": 30},
        {"name": "SMA", "epoch": 10, "pop_size": 30}
    ]
    list_obj = ["BHI", "MIS", "XBI"]

    model = MetaCluster(list_optimizer=list_optimizer, list_paras=list_paras, list_obj=list_obj, n_trials=3)
    assert model.n_trials == 3
    assert model.list_obj == list_obj
    assert len(model.list_optimizer) == len(list_optimizer)



#!/usr/bin/env python
# Created by "Thieu" at 23:33, 21/05/2022 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

import pandas as pd
import numpy as np
from pathlib import Path
from sklearn.model_selection import train_test_split
from sklearn import preprocessing


class Data:
    """
    The structure of our supported Data class

    Parameters
    ----------
    X : np.ndarray
        The features of your data

    y : np.ndarray, Optional, default=None
        The labels of your data, for clustering problem, this can be None
    """

    SUPPORT = {
        "scaler": ["StandardScaler", "MinMaxScaler", "MaxAbsScaler", "RobustScaler", "Normalizer"]
    }

    def __init__(self, X, y=None, name="Unknown"):
        self.X = X
        self.y = y
        self.name = name
        self.X_train, self.y_train, self.X_test, self.y_test = None, None, None, None

    def split_train_test(self, test_size=0.2, train_size=None,
                         random_state=41, shuffle=True, stratify=None, inplace=True):
        """
        The wrapper of the split_train_test function in scikit-learn library.
        """
        if self.y is None:
            self.X_train, self.X_test = train_test_split(self.X, test_size=test_size,
                        train_size=train_size, random_state=random_state, shuffle=shuffle, stratify=stratify)
        else:
            self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(self.X, self.y, test_size=test_size,
                        train_size=train_size, random_state=random_state, shuffle=shuffle, stratify=stratify)
        if not inplace:
            return self.X_train, self.X_test, self.y_train, self.y_test

    @staticmethod
    def scale(X, method="MinMaxScaler", **kwargs):
        if method in Data.SUPPORT["scaler"]:
            scaler = getattr(preprocessing, method)(**kwargs)
            data = scaler.fit_transform(X)
            return data, scaler
        raise ValueError(f"Data class doesn't support scaling method name: {method}")

    def set_train_test(self, X_train=None, y_train=None, X_test=None, y_test=None):
        """
        Function use to set your own X_train, y_train, X_test, y_test in case you don't want to use our split function

        Parameters
        ----------
        X_train : np.ndarray
        y_train : np.ndarray
        X_test : np.ndarray
        y_test : np.ndarray
        """
        self.X_train = X_train
        self.y_train = y_train
        self.X_test = X_test
        self.y_test = y_test

    def get_name(self):
        return self.name


def get_dataset(dataset_name):
    """
    Helper function to retrieve the data

    Parameters
    ----------
    dataset_name : str
        Name of the dataset

    Returns
    -------
    data: Data
        The instance of Data class, that hold X and y variables.
    """
    dir_root = f"{Path(__file__).parent.parent.__str__()}/data"
    list_path = Path(f"{dir_root}").glob("*.csv")
    list_datasets = [pf.name[:-4] for pf in list_path]

    if dataset_name not in list_datasets:
        print(f"MetaCluster currently does not have '{dataset_name}' data in its database....")
        print("+ List of the supported datasets are:")
        for idx, dataset in enumerate(list_datasets):
            print(f"\t{idx + 1}: {dataset}")
    else:
        df = pd.read_csv(f"{dir_root}/{dataset_name}.csv", header=None)
        data = Data(np.array(df.iloc[:, 0:-1]), np.array(df.iloc[:, -1]), name=dataset_name)
        print(f"Requested dataset: {dataset_name} found and loaded!")
        return data



#!/usr/bin/env python
# Created by "Thieu" at 14:52, 26/05/2023 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

import numpy as np
from mealpy.optimizer import Optimizer
from mealpy import *
import sys, inspect
from permetrics import ClusteringMetric


EXCLUDE_MODULES = ["__builtins__", "current_module", "inspect", "sys"]


def get_all_optimizers():
    cls = {}
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.ismodule(obj) and (name not in EXCLUDE_MODULES):
            for cls_name, cls_obj in inspect.getmembers(obj):
                if inspect.isclass(cls_obj) and issubclass(cls_obj, Optimizer):
                    cls[cls_name] = cls_obj
    del cls['Optimizer']
    return cls


def get_optimizer_by_name(name):
    try:
        cls = get_all_optimizers()[name]
        return cls
    except KeyError:
        print(f"MetaCluster doesn't support optimizer named: {name}.\n"
              f"Please see the supported Optimizer name from here: https://mealpy.readthedocs.io/en/latest/pages/support.html#classification-table")


class KCenterClusteringProblem(Problem):
    def __init__(self, lb, ub, minmax, data=None, estimator=None, obj_name=None, obj_paras=None,
                 name="K Center Clustering Problem", **kwargs):
        ## data is assigned first because when initialize the Problem class, we need to check the output of fitness
        self.data = data
        self.estimator = estimator
        self.obj_name = obj_name
        self.obj_paras = obj_paras
        self.name = name
        super().__init__(lb, ub, minmax, **kwargs)

    def get_y_pred(self, X, solution):
        centers = np.reshape(solution, (-1, X.shape[1]))
        # Calculate the distance between each sample and each center
        distances = np.sqrt(np.sum((X[:, np.newaxis, :] - centers) ** 2, axis=2))
        # Assign each sample to the closest center
        labels = np.argmin(distances, axis=1)
        return labels

    def get_metrics(self, solution=None, list_metric=None, list_paras=None):
        centers = np.reshape(solution, (-1, self.data.X.shape[1]))
        y_pred = self.get_y_pred(self.data.X, centers)
        evaluator = ClusteringMetric(y_true=self.data.y, y_pred=y_pred, X=self.data.X)
        results = evaluator.get_metrics_by_list_names(list_metric, list_paras)
        return results

    # def amend_position(self, position=None, lb=None, ub=None):
    #     n_features = self.data.X.shape[1]
    #     n_clusters = int(len(position) / n_features)
    #     pos = np.clip(position, lb, ub)
    #     centers = np.reshape(pos, (n_clusters, n_features))
    #     y_pred = self.get_y_pred(self.data.X, centers)
    #     while len(np.unique(y_pred, return_counts=True)[0]) == 1:
    #         centers[np.random.randint(0, n_clusters)] = np.random.uniform(lb[:n_features], ub[:n_features])
    #         y_pred = self.get_y_pred(self.data.X, centers)
    #     return centers.flatten()

    def fit_func(self, solution):
        centers = np.reshape(solution, (-1, self.data.X.shape[1]))
        y_pred = self.get_y_pred(self.data.X, centers)
        evaluator = ClusteringMetric(y_true=self.data.y, y_pred=y_pred, X=self.data.X, raise_error=False)
        obj = evaluator.get_metric_by_name(self.obj_name, paras=self.obj_paras)[self.obj_name]
        return obj



#!/usr/bin/env python
# Created by "Thieu" at 13:24, 25/05/2023 ----------%
#       Email: nguyenthieu2102@gmail.com            %
#       Github: https://github.com/thieu1995        %
# --------------------------------------------------%

from setuptools import setup, find_packages


def readme():
    with open('README.md', encoding='utf-8') as f:
        README = f.read()
    return README


setup(
    name="metacluster",
    version="0.2.0",
    author="Thieu",
    author_email="nguyenthieu2102@gmail.com",
    description="MetaCluster: An Open-Source Python Library for Metaheuristic-based Clustering Problems",
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords=["clustering", "optimization", "k-center clustering", "data points", "centers", "euclidean distance", "maximum distance",
              "NP-hard", "greedy algorithm", "approximation algorithm", "covering problem", "computational complexity",
              "geometric algorithms", "machine learning", "pattern recognition", "spatial analysis", "graph theory", "mathematical optimization",
              "dimensionality reduction", "mutual information", "correlation-based feature selection",
              "Genetic algorithm (GA)", "Particle swarm optimization (PSO)", "Ant colony optimization (ACO)",
              "Differential evolution (DE)", "Simulated annealing", "Grey wolf optimizer (GWO)", "Whale Optimization Algorithm (WOA)",
              "confusion matrix", "recall", "precision", "accuracy", "K-Nearest Neighbors",
              "pearson correlation coefficient (PCC)", "spearman correlation coefficient (SCC)",
              "multi-objectives optimization problems", "Stochastic optimization", "Global optimization",
              "Convergence analysis", "Search space exploration", "Local search", "Computational intelligence", "Robust optimization",
              "Performance analysis", "Intelligent optimization", "Simulations"],
    url="https://github.com/thieu1995/metacluster",
    project_urls={
        'Documentation': 'https://metacluster.readthedocs.io/',
        'Source Code': 'https://github.com/thieu1995/metacluster',
        'Bug Tracker': 'https://github.com/thieu1995/metacluster/issues',
        'Change Log': 'https://github.com/thieu1995/metacluster/blob/master/ChangeLog.md',
        'Forum': 'https://t.me/+fRVCJGuGJg1mNDg1',
    },
    packages=find_packages(exclude=['tests*', 'examples*']),
    include_package_data=True,
    license="GPLv3",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: System :: Benchmark",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    install_requires=["numpy>=1.17.1", "scipy>=1.7.1", "scikit-learn>=1.0.2",
                      "pandas>=1.3.5", "mealpy>=2.5.3", "permetrics>=1.4.1",
                      "plotly>=5.10.0", "kaleido>=0.2.1"],
    extras_require={
        "dev": ["pytest>=7.1.2", "pytest-cov==4.0.0", "flake8>=4.0.1"],
    },
    python_requires='>=3.7',
)
